use rand::prelude::ThreadRng;
use rand::thread_rng;
use rand::Rng;

type Reine = (usize, usize);
type Individual = Vec<Reine>;
type Population = Vec<Individual>;

const NB_SWAPS_PER_INDIVIDUAL: usize = 100;

fn can_be_put(reine: Reine, board: &Individual) -> bool {
    for o in board {
        if reine.0 == o.0
            || reine.1 == o.1
            || (o.0 as isize - reine.0 as isize).abs() == (o.1 as isize - reine.1 as isize).abs()
        {
            return false;
        }
    }

    true
}

fn lambda_individual(dim: usize) -> Individual {
    let mut chessboard: Vec<Reine> = Vec::new();

    for i in 0..dim {
        chessboard.push((i, i));
    }

    chessboard
}

fn mutate(x: &mut Individual, rng: &mut ThreadRng) -> () {
    let first_index: usize = (rng.gen::<f64>() * x.len() as f64).abs() as usize;
    let mut second_index: usize = (rng.gen::<f64>() * x.len() as f64).abs() as usize;

    while second_index == first_index {
        second_index = (rng.gen::<f64>() * x.len() as f64).abs() as usize;
    }

    let tmp = x[first_index];

    x[first_index] = x[second_index];
    x[second_index] = tmp;
}

fn fitness(individual: &Individual) -> usize {
    let mut board: Individual = Vec::new();

    let mut i = 0;

    for reine in individual {
        if can_be_put(*reine, &board) {
            i += 1;
        }

        board.push(*reine);
    }

    i
}

pub fn nqueens_crossover(a: &Individual, b: &Individual) -> Individual {
    let mut rng: ThreadRng = rand::thread_rng();

    let mut crossover = b.clone();

    //so that locus start isn't on the last cell
    let locus_start = rng.gen_range(0, a.len() - 1);
    let locus_end = rng.gen_range(locus_start + 1, a.len());

    let locus = &a[locus_start..=locus_end];

    let locus_cols = locus.iter().map(|(_, col)| *col).collect::<Vec<usize>>();

    let mut locus_position = 0;

    for i in 0..crossover.len() {
        let (_, col) = crossover[i];

        if locus_position < locus.len() && locus_cols.contains(&col) {
            crossover[i].1 = locus[locus_position].1;
            locus_position += 1;
        }
    }

    crossover
}

pub fn generate_population(dim: usize, nb_individuals: usize, nb_swaps: usize) -> Population {
    let mut rng = thread_rng();

    let mut population = Vec::with_capacity(nb_individuals);

    for _ in 0..nb_individuals {
        population.push(generate_new_individual(dim, nb_swaps, &mut rng));
    }

    population
}

pub fn generate_new_individual(dim: usize, nb_swaps: usize, rng: &mut ThreadRng) -> Individual {
    let mut individual = lambda_individual(dim);

    for _ in 0..nb_swaps {
        mutate(&mut individual, rng);
    }

    individual
}

pub fn bias_pick_from(weighed_probability_map: &Vec<(usize, usize)>, rng: &mut ThreadRng) -> usize {
    let fitnesses_sum: usize = weighed_probability_map[weighed_probability_map.len() - 1].1;

    let roll: usize = rng.gen_range(0, fitnesses_sum + 1);

    //look for picked individual
    let mut ind_idx = 0;

    for (idx, fitness_acc) in weighed_probability_map {
        ind_idx = *idx;
        if *fitness_acc >= roll {
            break;
        }
    }

    ind_idx
}

pub fn nqueens_genetic(dim: usize, population_size: usize, verbose: bool) -> Option<Individual> {
    let mut population: Population =
        generate_population(dim, population_size, NB_SWAPS_PER_INDIVIDUAL);

    if verbose {
      println!("Running genetic algorithm with parameters:\n\tDimension: {}\n\tPopulation size: {}", dim, population_size);
    }

    let solution: Option<&Individual>;

    let mut pop_number = 1;

    loop {
        let fitness_map: Vec<(usize, usize)> = population
            .iter()
            .enumerate()
            .map(|(i, x)| (i, fitness(x)))
            .collect();

        if verbose {
          println!("{},{}",
            pop_number,
            fitness_map
              .iter()
              .max_by_key(|(_, fitness)| fitness)
              .unwrap()
              .1
          );
          
          pop_number+=1
        }

        //Condition d'arrêt de la boucle de recherche
        if let Some((idx, _)) = fitness_map.iter().find(|(_, fitness)| *fitness == dim) {
            solution = Some(&population[*idx]);
            break;
        }

        //Fait correspondre les indices des individus à leur score.
        //Leur score est utilisé comme poids pour un tirage au hasard biaisé
        let weighed_probability_map: Vec<(usize, usize)> = fitness_map
            .iter()
            .map(|(i, _)| {
                (
                    *i,
                    fitness_map[0..=*i].iter().map(|(_, fitness)| fitness).sum(),
                )
            })
            .collect();

        let mut rng = rand::thread_rng();

        let mut next_gen: Population = Vec::with_capacity(population_size);

        //On garde 15% des meilleurs, 5% des moins bons, et 75% d'enfants des meilleurs
        //On mélange à ça 5% nouveaux individus
        for _ in 0..(0.75 * population_size as f64) as usize {
            let ind1 = &population[bias_pick_from(&weighed_probability_map, &mut rng)];
            let mut ind2 = &population[bias_pick_from(&weighed_probability_map, &mut rng)];

            while ind1 == ind2 {
                ind2 = &population[bias_pick_from(&weighed_probability_map, &mut rng)];
            }

            next_gen.push(nqueens_crossover(&ind1, &ind2));
        }

        //Sorting the individuals by their fitness
        population.sort_by_key(|ind| fitness(ind));

        //15% best individuals
        let fifteen_percent_best = &mut population
            [(population_size - (0.15 * population_size as f64) as usize)..population_size]
            .to_vec();
        next_gen.append(fifteen_percent_best);

        //5% new individuals
        let mut new_individuals = generate_population(
            dim,
            (0.05 * population_size as f64) as usize,
            NB_SWAPS_PER_INDIVIDUAL,
        );
        next_gen.append(&mut new_individuals);

        //Rest filled with worst individuals
        let rest_count = population_size - next_gen.len();
        if rest_count > 0 {
            let five_percent_worst = &mut population[0..rest_count].to_vec();
            next_gen.append(five_percent_worst);
        }

        population = next_gen;
    }

    solution.map(|s| s.clone())
}