#[macro_use]
extern crate clap;
extern crate colored;
extern crate rand;

mod annealing;
mod backtrack;
mod genetic;
mod heuristics;

use std::time::SystemTime;

use clap::App;
use colored::*;

use backtrack::*;

use heuristics::HeuristicsApproach;

use annealing::{nqueens_annealing, test_annealing};

use genetic::nqueens_genetic;

fn main() {
    let yaml = load_yaml!("cli.yml");
    let matches = App::from_yaml(yaml).get_matches();

    let subcommand = matches.subcommand();

    match subcommand.0 {
        "backtrack" => backtrack_solution(subcommand.1.unwrap()),
        "annealing" => annealing_solution(subcommand.1.unwrap()),
        "genetic" => genetic_solution(subcommand.1.unwrap()),
        _ => panic!("unknown subcommand"),
    }
}

fn backtrack_solution(matches: &clap::ArgMatches<'_>) {
    let verbose = matches.is_present("verbose");
    let first = matches.is_present("first");
    let color = matches.is_present("color");

    let dim = matches.value_of("dimension");

    let series = matches.values_of("series").map(|v| {
        let values: Vec<_> = v.collect();
        values
    });

    if series.is_some() {
        let ser = series.unwrap();
        let min: usize = ser[0].parse().unwrap();
        let max: usize = ser[1].parse().unwrap();

        for n in min..=max {
            let start = SystemTime::now();

            let opportunist = matches.is_present("opportunist");
            let failing = matches.is_present("failing");

            let heuristics: Option<HeuristicsApproach> = match (opportunist, failing) {
                (true, _) => Some(HeuristicsApproach::Opportunist),
                (_, true) => Some(HeuristicsApproach::Failing),
                _ => None,
            };

            let found = match heuristics {
                Some(approach) => nreines_first_solution_with_heuristics(n, verbose, approach),
                None => nreines_first_solution(n, verbose),
            };

            let end = SystemTime::now();

            let duration = end.duration_since(start).unwrap();

            let mut n_string = n.to_string();

            let mut found_status = if found {
                "found".to_string()
            } else {
                "not found".to_string()
            };

            let mut elapsed =
                (duration.as_secs() * 1000 + u64::from(duration.subsec_millis())).to_string();

            if color {
                n_string = n_string.bold().to_string();

                found_status = if found {
                    found_status.blue().to_string()
                } else {
                    found_status.red().to_string()
                };

                elapsed = elapsed.green().to_string();
            }

            println!("{:0>2} {:>9} in {:<8}ms", n_string, found_status, elapsed,);
        }
    } else if dim.is_some() {
        let n: usize = dim.unwrap().parse().unwrap();

        let start = SystemTime::now();

        let mut results: (Option<bool>, Option<usize>) = (None, None);

        if first {
            let opportunist = matches.is_present("opportunist");
            let failing = matches.is_present("failing");
            let heuristics: Option<HeuristicsApproach> = match (opportunist, failing) {
                (true, _) => Some(HeuristicsApproach::Opportunist),
                (_, true) => Some(HeuristicsApproach::Failing),
                _ => None,
            };

            results.0 = Some(match heuristics {
                Some(approach) => nreines_first_solution_with_heuristics(n, verbose, approach),
                None => nreines_first_solution(n, verbose),
            });
        } else {
            results.1 = Some(nreines(n, verbose));
        }

        let end = SystemTime::now();

        let duration = end.duration_since(start).unwrap();

        if results.0.is_some() {
            let found = results.0.unwrap();

            let mut found_status = if found {
                "found".to_string()
            } else {
                "not found".to_string()
            };

            let mut elapsed =
                (duration.as_secs() * 1000 + u64::from(duration.subsec_millis())).to_string();

            if color {
                found_status = if found {
                    found_status.blue().to_string()
                } else {
                    found_status.red().to_string()
                };

                elapsed = elapsed.green().to_string();
            }

            println!("Solution {} in {:>8}ms", found_status, elapsed,);
        } else if results.1.is_some() {
            let mut n_string = results.1.unwrap().to_string();

            let mut elapsed =
                (duration.as_secs() * 1000 + u64::from(duration.subsec_millis())).to_string();

            if color {
                n_string = n_string.bold().to_string();
                elapsed = elapsed.green().to_string();
            }

            println!("{} solutions found in {:>8}ms", n_string, elapsed,);
        }
    }
}

fn annealing_solution(matches: &clap::ArgMatches<'_>) {
    let test = matches.is_present("test");
    let verbose = matches.is_present("verbose");

    if test {
        println!("{}", test_annealing());
    } else {
        let dimension: usize = matches.value_of("dimension").unwrap().parse().unwrap();

        let start = SystemTime::now();

        let result = nqueens_annealing(dimension, verbose);

        let end = SystemTime::now();

        if result.y == 0 {
            let duration = end.duration_since(start).unwrap();

            let elapsed = duration.as_secs() * 1000 + u64::from(duration.subsec_millis());

            println!("Solution found in {}ms:\n{:?}", elapsed, result.x);
        } else {
            println!("Solution was not found");
        }
    }
}

fn genetic_solution(matches: &clap::ArgMatches<'_>) {
    let verbose = matches.is_present("verbose");

    let dimension: usize = matches.value_of("dimension").unwrap().parse().unwrap();
    let population_size: usize = matches.value_of("population").unwrap().parse().unwrap();

    let start = SystemTime::now();

    let result = nqueens_genetic(dimension, population_size, verbose);

    let end = SystemTime::now();

    if result.is_some() {
        let duration = end.duration_since(start).unwrap();

        let elapsed = duration.as_secs() * 1000 + u64::from(duration.subsec_millis());

        println!("Solution found in {}ms:\n{:?}", elapsed, result.unwrap());
    } else {
        println!("Solution was not found");
    }
}